// WorldObject.cpp: implementation of the WorldObject class.
//
//////////////////////////////////////////////////////////////////////

#include "WorldObject.h"
#include <stdio.h>
#include <stdlib.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

WorldObject::WorldObject()
{
	position.SetXYZ(0.0,0.0,0.0);
	velocity.SetXYZ(0.0,0.0,0.0);
}

WorldObject::WorldObject(PointVector pos,float objectmass,float objectsize)
{
	position.SetXYZ(pos.GetX(),pos.GetY(),pos.GetZ());
	velocity.SetXYZ(0.0,0.0,0.0);
	mass = objectmass;
	size = objectsize;
}

WorldObject::WorldObject(const WorldObject& anotherobject)
{
	position = anotherobject.position;
	velocity = anotherobject.velocity;
	size = anotherobject.size;
	mass = anotherobject.mass;
	idnumber = anotherobject.idnumber;
}

WorldObject::~WorldObject()
{

}

void WorldObject::SetPosition(PointVector pos)
{
	position.SetXYZ(pos.GetX(),pos.GetY(),pos.GetZ());
}

void WorldObject::Accelerate(PointVector accelerationVector)
{
	velocity.Add(accelerationVector.GetX(),accelerationVector.GetY(),accelerationVector.GetZ());
}

void WorldObject::Move()
{
	position.Add(velocity.GetX(),velocity.GetY(),velocity.GetZ());
}

void WorldObject::SetVelocity(PointVector vel)
{
	velocity.SetXYZ(vel.GetX(),vel.GetY(),vel.GetZ());
}

PointVector WorldObject::GetPosition()
{
	return position;
}

PointVector WorldObject::GetVelocity()
{
	return velocity;
}

bool WorldObject::SphereCollision(WorldObject anotherObject)
{
	// distance between this object's center point and anotherObject's center point
	float distance = position.DistanceBetweenPoint(anotherObject.position);
	// totalradius is the distance between two sphere objects when hit
	float totalradius = size + anotherObject.size;

	if(distance <= totalradius)
		return true;	// collision is true
	else
		return false;	// no collision
}

bool WorldObject::GroundCollision()
{
	// distance between this object's center point and the ground y = 0
	float distance = position.GetY();
	if(distance <= size)
		return true;
	else
		return false;
}


void WorldObject::SetSize(float setsize)
{
	size = setsize;
}

const WorldObject& WorldObject::operator =(const WorldObject &someObject)
{
	if(&someObject != this)
	{
		position = someObject.position;
		velocity = someObject.velocity;
		size = someObject.size;
		mass = someObject.mass;
		idnumber = someObject.idnumber;
	}
	return *this;
}


PointVector WorldObject::ReflectOffObject(WorldObject otherObject,float elasticity)
{
	// if position and size of the two objects are the same nothing happens
	if(!(otherObject.GetPosition()).Equals(position) && otherObject.GetSize() != size)
	{	
		// this is to avoid an object from reflecting off itself
		// 2n(n.l) - l = r
		float collisionvelocity = velocity.Length();
		float energy = 0.5 * mass * collisionvelocity;

		PointVector normal = position.SubtractVector(otherObject.position);
		normal.Normalize();
	
		PointVector normalizedVelocity = velocity;
		normalizedVelocity.Normalize();
		normalizedVelocity.Reverse();

		float dotProduct = 2.0 * normal.DotProduct(normalizedVelocity);
		normal.MultiplyScalar(dotProduct);
	
		PointVector returnVector = normal.SubtractVector(normalizedVelocity);
	
		float reflectedvelocity = returnVector.Length();
		float scale = (collisionvelocity / reflectedvelocity) * elasticity;
		returnVector.MultiplyScalar(scale * 2.0);
		/*
		// if dead on from top, nudge slightly to x and z
		if(velocity.XIsZero() && velocity.ZIsZero())
		{
			float addxz = 0.000001;
			float xsign = 1.0;
			float zsign = 1.0;
			int choicex = rand() % 2;
			int choicez = rand() % 2;
			if(choicex == 1)
				xsign = -1.0;
			else
				xsign = 1.0;
			if(choicez == 1)
				zsign = -1.0;
			else
				zsign = 1.0;
			returnVector.Add(xsign * addxz,0.0,zsign * addxz);
		}
		*/		
		return returnVector;
	}
	PointVector zerovector;
	return zerovector;
}

PointVector WorldObject::ReflectOffGround(float elasticity)
{
	// 2n(n.l) - l = r
	//printf("*************************************\n");
	float collisionvelocity = velocity.Length();
	float energy = 0.5 * mass * collisionvelocity;
	//printf("CollisionVelocity: %f\n",collisionvelocity);
	//printf("Energy: %f\n",energy);

	PointVector normal = PointVector(0.0,1.0,0.0);

	PointVector normalizedVelocity = velocity;
	//normalizedVelocity.Normalize();
	normalizedVelocity.Reverse();

	float dotProduct = 2.0 * normal.DotProduct(normalizedVelocity);
	normal.MultiplyScalar(dotProduct);

	PointVector returnVector = normal.SubtractVector(normalizedVelocity);
	//printf("ReflectedVelocityVector: ");
	//returnVector.Print();
	
	float reflectedvelocity = returnVector.Length();
	//printf("ReflectedVelocity: %f\n",reflectedvelocity);
	float scale = (collisionvelocity / reflectedvelocity) * elasticity;
	//printf("Scale: %f\n",scale);
	returnVector.MultiplyScalar(scale * 2.0);
	// if dead on from top, nudge slightly to x and z
	/*
	if(velocity.XIsZero() && velocity.ZIsZero())
	{
		float addxz = 0.000001;
		float xsign = 1.0;
		float zsign = 1.0;
		int choicex = rand() % 2;
		int choicez = rand() % 2;
		if(choicex == 1)
			xsign = -1.0;
		else
			xsign = 1.0;
		if(choicez == 1)
			zsign = -1.0;
		else
			zsign = 1.0;
		returnVector.Add(xsign * addxz,0.0,zsign * addxz);
	}
	*/
	return returnVector;
}


void WorldObject::Print()
{
	printf("********************\n");
	printf("Position :");
	position.Print();
	printf("Velocity :");
	velocity.Print();
}

void WorldObject::SetMass(float objectmass)
{
	mass = objectmass;
}

void WorldObject::MoveOutOf(WorldObject obj)
{
	if(SphereCollision(obj))
	{
		// nudge out in normal direction of obj's surface
		PointVector normal = position.SubtractVector(obj.position);
		float properdistance = size + obj.size;
		float actualdistance = position.DistanceBetweenPoint(obj.position);
		// amount need to nudge out is properdistance - actualdistance
		float amount = properdistance - actualdistance;
		amount = amount / properdistance;
		// this is the ration we multiply our normal
		normal.MultiplyScalar(amount);
		// we move our object by this normal
		Move(normal);
	}
}

void WorldObject::Move(PointVector distance)
{
	position.Add(distance.GetX(),distance.GetY(),distance.GetX());
}

void WorldObject::ResetSpeedPosition()
{
	position.Reset();
	velocity.Reset();
}

void WorldObject::ResetSpeed()
{
	velocity.Reset();
}

void WorldObject::SetID(int id)
{
	idnumber = id;
}

int WorldObject::GetID()
{
	return idnumber;
}

float WorldObject::GetSize()
{
	return size;
}

void WorldObject::MoveAboveGround()
{
	if(GroundCollision())
	{
		position.SetY(size);
	}
}

void WorldObject::SphereCollisionRoutine(WorldObject otherObject,float elasticity)
{
	PointVector reflectVector = ReflectOffObject(otherObject,elasticity);
	MoveOutOf(otherObject);
	Accelerate(reflectVector);
}

float WorldObject::GroundCollisionRoutine(float elasticity)
{
	PointVector reflectVector = ReflectOffGround(elasticity);
	MoveAboveGround();
	velocity.Reset();
	//Accelerate(reflectVector);
	reflectVector.MultiplyScalar(elasticity);
	//reflectVector.Reverse();
	Accelerate(reflectVector);
	float returnenergy = 0.5 * mass * velocity.Length();
	return returnenergy;
}

bool WorldObject::BelowGround()
{
	if(position.GetY() < 0.001)
		return true;
	else
		return false;
}
